from __future__ import annotations

import holoviews as hv

bk_renderer = hv.renderer("bokeh")


def _get_display(annotator, kdims=None) -> hv.Element:
    if kdims is None:
        kdims = next(iter(annotator._displays))
    kdims = (kdims,) if isinstance(kdims, str) else tuple(kdims)
    return annotator.get_display(*kdims)


def get_region_editor_data(
    annotator,
    element_type,
    kdims=None,
):
    el = _get_display(annotator, kdims).region_editor()
    for e in el.last.traverse():
        if isinstance(e, element_type):
            return e.data


def get_indicators_data(annotator, element_type, kdims=None):
    return _get_display(annotator, kdims).static_indicators.data


def test_set_regions_range1d(annotator_range1d, element_range1d) -> None:
    annotator = annotator_range1d
    element = element_range1d
    annotator_element = annotator * element
    bk_renderer.get_plot(annotator_element)

    # No regions has been set
    output = get_region_editor_data(annotator, hv.VSpan)
    expected = [None, None]
    assert output == expected

    # Setting regions
    annotator.set_regions(TIME=(-0.25, 0.25))
    output = get_region_editor_data(annotator, hv.VSpan)
    expected = [-0.25, 0.25]
    assert output == expected

    # Adding annotation and remove selection regions.
    annotator.add_annotation(description="Test")
    output = get_region_editor_data(annotator, hv.VSpan)
    expected = [None, None]
    assert output == expected

    output = get_indicators_data(annotator, hv.Rectangles)
    output1 = output.loc[0, ["start[TIME]", "end[TIME]"]].tolist()
    expected1 = [-0.25, 0.25]
    assert output1 == expected1
    output2 = output.loc[0, "description"]
    expected2 = "Test"
    assert output2 == expected2


def test_set_regions_range2d(annotator_range2d, element_range2d) -> None:
    annotator = annotator_range2d
    element = element_range2d
    annotator_element = annotator * element
    bk_renderer.get_plot(annotator_element)

    # No regions has been set
    output = get_region_editor_data(annotator, hv.Rectangles)
    assert output.empty

    # Setting regions
    annotator.set_regions(x=(-0.25, 0.25), y=(-0.25, 0.25))
    output = get_region_editor_data(annotator, hv.Rectangles).iloc[0].to_list()
    expected = [-0.25, -0.25, 0.25, 0.25]
    assert output == expected

    # Adding annotation and remove selection regions.
    annotator.add_annotation(description="Test")
    output = get_region_editor_data(annotator, hv.Rectangles)
    assert output.empty

    output = get_indicators_data(annotator, hv.Rectangles)
    output1 = output.loc[0, ["start[x]", "start[y]", "end[x]", "end[y]"]].tolist()
    expected1 = [-0.25, -0.25, 0.25, 0.25]
    assert output1 == expected1
    output2 = output.loc[0, "description"]
    expected2 = "Test"
    assert output2 == expected2


def test_set_regions_multiple(multiple_annotators, element_range1d, element_range2d):
    annotator = multiple_annotators
    time_annotation = annotator * element_range1d
    xy_annotation = annotator * element_range2d
    annotator_element = time_annotation + xy_annotation
    bk_renderer.get_plot(annotator_element)

    # No regions has been set
    # Time annotation
    output = get_region_editor_data(annotator, hv.VSpan, "TIME")
    expected = [None, None]
    assert output == expected
    # xy annotation
    output = get_region_editor_data(annotator, hv.Rectangles, ("x", "y"))
    assert output.empty

    # Setting regions
    annotator.set_regions(TIME=(-0.25, 0.25), x=(-0.25, 0.25), y=(-0.25, 0.25))
    # Time annotation
    output = get_region_editor_data(annotator, hv.VSpan, "TIME")
    expected = [-0.25, 0.25]
    assert output == expected
    # xy annotation
    output = get_region_editor_data(annotator, hv.Rectangles, ("x", "y")).iloc[0].to_list()
    expected = [-0.25, -0.25, 0.25, 0.25]
    assert output == expected

    # Adding annotation and remove selection regions.
    annotator.add_annotation(description="Test")
    # Time annotation
    output = get_region_editor_data(annotator, hv.VSpan, "TIME")
    expected = [None, None]
    assert output == expected

    output = get_indicators_data(annotator, hv.Rectangles, "TIME")
    output1 = output.loc[0, ["start[TIME]", "end[TIME]"]].tolist()
    expected1 = [-0.25, 0.25]
    assert output1 == expected1
    output2 = output.loc[0, "description"]
    expected2 = "Test"
    assert output2 == expected2

    # xy annotation
    output = get_region_editor_data(annotator, hv.Rectangles, ("x", "y"))
    assert output.empty

    output = get_indicators_data(annotator, hv.Rectangles, ("x", "y"))
    output1 = output.loc[0, ["start[x]", "start[y]", "end[x]", "end[y]"]].tolist()
    expected1 = [-0.25, -0.25, 0.25, 0.25]
    assert output1 == expected1
    output2 = output.loc[0, "description"]
    expected2 = "Test"
    assert output2 == expected2


def test_editable_enabled(annotator_range1d, element_range1d):
    annotator_range1d * element_range1d

    annotator_range1d.editable_enabled = False
    for display in annotator_range1d._displays.values():
        assert not display.editable_enabled

    annotator_range1d.editable_enabled = True
    for display in annotator_range1d._displays.values():
        assert display.editable_enabled


def test_selection_enabled(annotator_range1d, element_range1d):
    annotator_range1d * element_range1d

    annotator_range1d.selection_enabled = False
    for display in annotator_range1d._displays.values():
        assert not display.selection_enabled

    annotator_range1d.selection_enabled = True
    for display in annotator_range1d._displays.values():
        assert display.selection_enabled



from __future__ import annotations

import contextlib
import datetime as dt
from typing import TYPE_CHECKING, Any

import panel as pn
import param

if TYPE_CHECKING:
    from holonote.annotate import Annotator


class PanelWidgets:
    mapping = {
        str: pn.widgets.TextInput,
        bool: pn.widgets.Checkbox,
        dt.datetime: pn.widgets.DatePicker,
        dt.date: pn.widgets.DatePicker,
        int: pn.widgets.IntSlider,
        float: pn.widgets.FloatSlider,
    }

    def __init__(self, annotator: Annotator, field_values: dict[str, Any] | None = None):
        self.annotator = annotator
        self.annotator.snapshot()
        self._widget_mode_group = pn.widgets.RadioButtonGroup(
            name="Mode", options=["+", "-", "✏"], width=90
        )
        self._widget_apply_button = pn.widgets.Button(name="✓", width=20)
        self._widget_revert_button = pn.widgets.Button(name="↺", width=20)
        self._widget_commit_button = pn.widgets.Button(name="▲", width=20)

        if field_values is None:
            self._fields_values = {k: "" for k in self.annotator.fields}
        else:
            self._fields_values = {k: field_values.get(k, "") for k in self.annotator.fields}
        self._fields_widgets = self._create_fields_widgets(self._fields_values)

        self._set_standard_callbacks()

    @property
    def tool_widgets(self):
        return pn.Row(
            self._widget_apply_button,
            pn.Spacer(width=10),
            self._widget_mode_group,
            pn.Spacer(width=10),
            self._widget_revert_button,
            self._widget_commit_button,
        )

    def _create_fields_widgets(self, fields_values):
        fields_widgets = {}
        for widget_name, default in fields_values.items():
            if isinstance(default, param.Parameter):
                parameterized = type("widgets", (param.Parameterized,), {widget_name: default})
                pane = pn.Param(parameterized)
                fields_widgets[widget_name] = pane.layout[1]
            elif isinstance(default, list):
                fields_widgets[widget_name] = pn.widgets.Select(
                    value=default[0], options=default, name=widget_name
                )
            else:
                widget_type = self.mapping[type(default)]
                if issubclass(widget_type, pn.widgets.TextInput):
                    fields_widgets[widget_name] = widget_type(
                        value=default, placeholder=widget_name, name=widget_name
                    )
                else:
                    fields_widgets[widget_name] = widget_type(value=default, name=widget_name)
        return fields_widgets

    @property
    def fields_widgets(self):
        accordion = False  # Experimental
        widgets = pn.Column(*self._fields_widgets.values())
        if accordion:
            return pn.Accordion(("fields", widgets))
        else:
            return widgets

    def _reset_fields_widgets(self):
        for widget_name, default in self._fields_values.items():
            if isinstance(default, param.Parameter):
                default = default.default
            with contextlib.suppress(Exception):
                # TODO: Fix when lists (for categories, not the same as the default!)
                self._fields_widgets[widget_name].value = default

    def _callback_apply(self, event):
        selected_ind = (
            self.annotator.selected_indices[0]
            if len(self.annotator.selected_indices) == 1
            else None
        )
        self.annotator.select_by_index()

        if self._widget_mode_group.value in ["+", "✏"]:
            fields_values = {k: v.value for k, v in self._fields_widgets.items()}
            if self._widget_mode_group.value == "+":
                self.annotator.add_annotation(**fields_values)
                self._reset_fields_widgets()
            elif (self._widget_mode_group.value == "✏") and (selected_ind is not None):
                self.annotator.update_annotation_fields(
                    selected_ind, **fields_values
                )  # TODO: Handle only changed
        elif self._widget_mode_group.value == "-" and selected_ind is not None:
            self.annotator.delete_annotation(selected_ind)

    def _callback_commit(self, event):
        self.annotator.commit()

    def _watcher_selected_indices(self, event):
        if len(event.new) != 1:
            return
        selected_index = event.new[0]
        # if self._widget_mode_group.value == '✏':
        for name, widget in self._fields_widgets.items():
            value = self.annotator.annotation_table._field_df.loc[selected_index][name]
            widget.value = value

    def _watcher_mode_group(self, event):
        if event.new in ["-", "✏"]:
            self.annotator.selection_enabled = True
            self.annotator.select_by_index()
            self.annotator.editable_enabled = False
        elif event.new == "+":
            self.annotator.editable_enabled = True
            self.annotator.select_by_index()
            self.annotator.selection_enabled = False

        for widget in self._fields_widgets.values():
            widget.disabled = event.new == "-"

    def _set_standard_callbacks(self):
        self._widget_apply_button.on_click(self._callback_apply)
        self._widget_revert_button.on_click(lambda event: self.annotator.revert_to_snapshot())
        self._widget_commit_button.on_click(self._callback_commit)
        self.annotator.param.watch(self._watcher_selected_indices, "selected_indices")
        self._widget_mode_group.param.watch(self._watcher_mode_group, "value")

    def __panel__(self):
        return pn.Column(self.fields_widgets, self.tool_widgets)



from __future__ import annotations

import contextlib
from collections.abc import Iterator

import holoviews as hv
import numpy as np
import pandas as pd
import pytest

from holonote.annotate import Annotator, SQLiteDB, UUIDHexStringKey


@pytest.fixture()
def conn_sqlite_uuid(tmp_path) -> Iterator[SQLiteDB]:
    conn = SQLiteDB(filename=str(tmp_path / "test.db"), primary_key=UUIDHexStringKey())
    yield conn
    with contextlib.suppress(Exception):
        conn.cursor.close()
    with contextlib.suppress(Exception):
        conn.con.close()


@pytest.fixture()
def annotator_range1d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"TIME": np.datetime64},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_point1d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"TIME": (np.datetime64, "point")},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_range2d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"x": float, "y": float},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_point2d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"x": (float, "point"), "y": (float, "point")},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def multiple_region_annotator(conn_sqlite_uuid) -> Annotator:
    return Annotator(
        {"TIME": np.datetime64},
        fields=["description"],
        # region_types=["Point", "Range"],
        connector=conn_sqlite_uuid,
    )


@pytest.fixture()
def multiple_annotators(conn_sqlite_uuid) -> Annotator:
    return Annotator(
        {"TIME": np.datetime64, "x": float, "y": float},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )


@pytest.fixture()
def multiple_fields_annotator(conn_sqlite_uuid) -> Annotator:
    conn_sqlite_uuid.fields = ["field1", "field2"]
    anno = Annotator(
        {"TIME": np.datetime64},
        fields=["field1", "field2"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def element_range1d() -> hv.Curve:
    time = pd.date_range("2020-01-01", "2020-01-10", freq="D").to_numpy()
    return hv.Curve(time, kdims=["TIME"])


@pytest.fixture()
def element_range2d() -> hv.Image:
    x = np.arange(10)
    xy = x[:, np.newaxis] * x
    return hv.Image(xy, kdims=["x", "y"])



from __future__ import annotations

import datetime as dt
from typing import TYPE_CHECKING, Any

import pandas as pd
import param

if TYPE_CHECKING:
    from .connector import Connector
    from .typing import SpecDict


class AnnotationTable(param.Parameterized):
    """
    Class that stores and manipulates annotation data, including methods
    to declare annotations and commit edits back to the original data
    source such as a database.
    """

    columns = ("region", "dim", "value", "_id")

    index = param.List(default=[])

    def __init__(self, **params):
        """
        Either specify annotation fields with filled field columns
        (via connector or dataframe) or declare the expected
        field columns if starting with no annotation data.
        """
        super().__init__(**params)

        self._region_df = pd.DataFrame(columns=self.columns)
        self._field_df = None

        self._edits = []
        self._index_mapping = {}

        self._update_index()
        self._field_df_snapshot, self._region_df_snapshot = None, None

    def load(self, connector=None, fields_df=None, primary_key_name=None, fields=None, spec=None):
        """
        Load the AnnotationTable from a connector or a fields DataFrame.
        """
        if fields is None:
            fields = []

        if [connector, primary_key_name] == [None, None]:
            msg = "Either a connector instance must be supplied or the primary key name supplied"
            raise ValueError(msg)
        if len(fields) < 1:
            msg = "More than one field column is required"
            raise ValueError(msg)
        primary_key_name = (
            primary_key_name if primary_key_name else connector.primary_key.field_name
        )

        if fields_df:
            fields_df = fields_df[fields].copy()  # Primary key/index for annotations
            self._field_df = fields_df
        elif connector:
            self.load_annotation_table(connector, fields, spec)
        elif fields_df is None:
            fields_df = pd.DataFrame(columns=[primary_key_name, *fields])
            fields_df = fields_df.set_index(primary_key_name)
            self._field_df = fields_df

        self.clear_edits()
        self._update_index()

    def update_annotation_region(self, region, index):
        value = region["value"]
        mask = self._region_df[self._region_df._id == index]
        assert len(mask) == 1, "TODO: Handle multiple region updates for single index"
        self._region_df.at[mask.index.values[0], "value"] = value
        self._edits.append(
            {"operation": "update", "id": index, "fields": None, "region_fields": []}
        )

    @property
    def has_snapshot(self) -> bool:
        return self._field_df_snapshot is not None

    def revert_to_snapshot(self):
        "Clears outstanding changes and used to implement an basic undo system."
        if self._field_df_snapshot is None:
            msg = "Call snapshot method before calling revert_to_snapshot"
            raise Exception(msg)
        self._field_df = self._field_df_snapshot
        self._region_df = self._region_df_snapshot
        self.clear_edits()

    def snapshot(self):
        "Saves a snapshot. Expected to only be used after a syncing commit"
        self._field_df_snapshot, self._region_df_snapshot = self._snapshot()

    def _snapshot(self):
        return self._field_df.copy(), self._region_df.copy()

    def _update_index(self) -> None:
        if self._field_df is None:
            self.index = []
            return

        self.index = list(self._field_df.index)

    def _expand_commit_by_id(self, id_val, fields=None, region_fields=None):
        kwargs = self._field_df.loc[[id_val]].to_dict("records")[0]
        if fields:
            kwargs = {k: v for k, v in kwargs.items() if k in fields}
        kwargs[self._field_df.index.name] = id_val
        if region_fields == []:
            return kwargs
        items = self._region_df[self._region_df["_id"] == id_val]
        for i in items.itertuples(index=False):
            if i.region == "range":
                kwargs[f"start_{i.dim}"] = i.value[0]
                kwargs[f"end_{i.dim}"] = i.value[1]
            else:
                kwargs[f"{i.region}_{i.dim}"] = i.value
        return kwargs

    def _expand_save_commits(self, ids):
        return {"field_list": [self._expand_commit_by_id(id_val) for id_val in ids]}

    def _create_commits(self):
        "Expands out the commit history into commit operations"
        fields_inds = set(self._field_df.index)
        region_inds = set(self._region_df["_id"].unique())
        unassigned_inds = fields_inds - region_inds
        if unassigned_inds:
            msg = f"Following annotations have no associated region: {unassigned_inds}"
            raise ValueError(msg)

        commits = []
        for edit in self._edits:
            operation = edit["operation"]
            if operation == "insert":
                # May be missing due to earlier deletion operation - nothing to do
                if edit["id"] not in self._field_df.index:
                    continue
                kwargs = self._expand_commit_by_id(edit["id"])

            elif operation == "delete":
                kwargs = {"id_val": edit["id"]}
            elif operation == "update":
                if edit["id"] not in self._field_df.index:
                    continue
                kwargs = self._expand_commit_by_id(
                    edit["id"], fields=edit["fields"], region_fields=edit["region_fields"]
                )
            elif operation == "save":
                kwargs = self._expand_save_commits(edit["ids"])
            commits.append({"operation": operation, "kwargs": kwargs})

        return commits

    def commits(self, connector):
        commits = self._create_commits()
        for commit in commits:
            operation = commit["operation"]
            kwargs = connector.transforms[operation](commit["kwargs"])
            getattr(connector, connector.operation_mapping[operation])(**kwargs)

        self.clear_edits()
        return commits

    def clear_edits(self, edit_type=None):
        "Clear edit state and index mapping"
        self._edits = []
        self._index_mapping = {}

    def add_annotation(self, regions: dict[str, Any], spec: SpecDict, **fields):
        "Takes a list of regions or the special value 'annotation-regions' to use attached annotators"
        index_value = fields.pop(self._field_df.index.name)
        self._add_annotation_fields(index_value, fields=fields)

        data = []
        for kdim, value in regions.items():
            if not value:
                continue

            d = {"region": spec[kdim]["region"], "dim": kdim, "value": value, "_id": index_value}
            data.append(
                # pd.DataFrame(d) does not work because tuples is expanded into multiple rows:
                # pd.DataFrame({'v': (1, 2)})
                pd.DataFrame(d.values(), index=d.keys()).T
            )

        self._region_df = pd.concat((self._region_df, *data), ignore_index=True)

        self._edits.append({"operation": "insert", "id": index_value})
        self._update_index()

    def _add_annotation_fields(self, index_value, fields=None):
        index_name = self._field_df.index.name
        index_name_set = set() if index_name is None else {index_name}
        unknown_kwargs = set(fields.keys()) - set(self._field_df.columns)
        if unknown_kwargs - index_name_set:
            unknown_str = ", ".join([f"{k!r}" for k in sorted(unknown_kwargs)])
            msg = f"Unknown fields columns: {unknown_str}"
            raise KeyError(msg)

        new_fields = pd.DataFrame([dict(fields, **{index_name: index_value})])
        new_fields = new_fields.set_index(index_name)
        if self._field_df.empty:
            self._field_df[new_fields.columns] = new_fields
        else:
            self._field_df = pd.concat((self._field_df, new_fields))

    def delete_annotation(self, index):
        if index is None:
            msg = "Deletion index cannot be None"
            raise ValueError(msg)
        self._region_df = self._region_df[
            self._region_df["_id"] != index
        ]  # Could match multiple rows
        self._field_df = self._field_df.drop(index, axis=0)

        self._edits.append({"operation": "delete", "id": index})
        self._update_index()

    def update_annotation_fields(self, index, **fields):
        for column, value in fields.items():
            self._field_df.loc[index, column] = value

        self._edits.append(
            {"operation": "update", "id": index, "fields": list(fields), "region_fields": []}
        )

    def define_fields(self, fields_df, index_mapping):
        # Need a staging area to hold everything till initialized
        self._index_mapping.update(index_mapping)  # Rename _field_df
        self._field_df = pd.concat([self._field_df, fields_df])
        self._edits.append({"operation": "save", "ids": list(fields_df.index)})

    def _empty_expanded_region_df(self, *, spec: SpecDict, dims: list[str] | None) -> pd.DataFrame:
        invalid_dims = set(dims) - spec.keys()
        if invalid_dims:
            invalid_dims_str = ", ".join([f"{dim!r}" for dim in sorted(invalid_dims)])
            msg = f"Dimension(s) {invalid_dims_str} not in the spec"
            raise ValueError(msg)

        columns, types = [], []
        for dim in dims:
            region = spec[dim]["region"]
            dtype = pd.NaT if issubclass(t := spec[dim]["type"], dt.date) else t()
            if region == "range":
                columns.extend([f"start[{dim}]", f"end[{dim}]"])
                types.extend([dtype, dtype])
            else:
                columns.append(f"{region}[{dim}]")
                types.append(dtype)

        return pd.DataFrame([types], columns=columns).drop(index=0)

    def _expand_region_df(self, *, spec: SpecDict, dims: list[str] | None = None) -> pd.DataFrame:
        data = self._region_df.pivot(index="_id", columns="dim", values="value")
        dims = list(dims or spec)

        expanded = self._empty_expanded_region_df(spec=spec, dims=dims)
        if data.empty:
            return expanded

        set_index = True
        for dim in dims:
            region = spec[dim]["region"]
            if dim not in data.columns:
                continue
            elif region == "range":
                na_mask = data[dim].isnull()
                data.loc[na_mask, dim] = data.loc[na_mask, dim].apply(lambda *x: (None, None))
                expanded[[f"start[{dim}]", f"end[{dim}]"]] = list(data[dim])
            else:
                dtype = expanded.dtypes[f"{region}[{dim}]"]
                expanded[f"{region}[{dim}]"] = data[dim].astype(dtype)

            if set_index:
                expanded.index = data.index
                set_index = False

        return expanded

    def get_dataframe(
        self, *, spec: SpecDict | None = None, dims: list[str] | None = None
    ) -> pd.DataFrame:
        field_df = self._field_df
        region_df = self._expand_region_df(spec=spec, dims=dims)

        df = region_df.merge(field_df, left_index=True, right_index=True, how="left")
        df.index.name = self._field_df.index.name
        df = df.reindex(field_df.index)
        return df

    def load_annotation_table(self, conn: Connector, fields: list[str], spec: SpecDict) -> None:
        """Load the AnnotationTable region and field DataFrame from a connector.

        Parameters
        ----------
        conn : Connector
            Database connection
        fields : list[str]
            List of field columns to load from the connector
        spec : SpecDict
            Dictionary of region specifications
        """
        df = conn.transforms["load"](conn.load_dataframe())

        # Load fields dataframe
        fields_df = df[fields].copy()
        self.define_fields(fields_df, {ind: ind for ind in fields_df.index})
        # Replace: self._field_df = pd.concat([self._field_df, fields_df])

        # Load region dataframe
        if df.empty:
            self._region_df = pd.DataFrame(columns=self.columns)
            return

        data = []
        for kdim, info in spec.items():
            region = info["region"]
            if region == "range":
                value = list(zip(df[f"start_{kdim}"], df[f"end_{kdim}"]))
            else:
                value = df[f"{region}_{kdim}"]

            subdata = pd.DataFrame(
                {"region": region, "dim": kdim, "value": value, "_id": list(df.index)}
            )
            if region == "range":
                empty_mask = subdata["value"] == (None, None)
            else:
                empty_mask = subdata["value"].isnull()

            data.append(subdata[~empty_mask])

        self._region_df = pd.concat(data, ignore_index=True)

        self._update_index()
        self.clear_edits()



from __future__ import annotations

import panel as pn

from holonote.app import PanelWidgets


def test_panel_app(annotator_range1d):
    w = PanelWidgets(annotator_range1d)
    assert isinstance(w.fields_widgets, pn.Column)
    assert isinstance(w.tool_widgets, pn.Row)



from __future__ import annotations

import warnings

from .annotator import Annotator  # noqa: F401
from .connector import (  # noqa: F401
    AutoIncrementKey,
    Connector,
    SQLiteDB,
    UUIDBinaryKey,
    UUIDHexStringKey,
)
from .table import *

# Ignore Bokeh UserWarning about multiple competiting tools introduced in 3.2.2
warnings.filterwarnings("ignore", category=UserWarning, module="holoviews.plotting.bokeh.plot")
del warnings



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from holonote.annotate import (
    AutoIncrementKey,
    Connector,
    UUIDBinaryKey,
    UUIDHexStringKey,
)


@pytest.fixture(params=[UUIDHexStringKey, AutoIncrementKey, UUIDBinaryKey])
def database(conn_sqlite_uuid, request):
    # Change the primary key type
    conn_sqlite_uuid.primary_key = request.param(field_name="uuid")
    fields = {
        "uuid": conn_sqlite_uuid.primary_key.schema,
        "description": "TEXT",
        "start": "TIMESTAMP",
        "end": "TIMESTAMP",
    }
    conn_sqlite_uuid.initialize(fields)
    return conn_sqlite_uuid


class TestConnector:
    "Tests for classmethods on the base class"

    def test_fields_from_metadata_literals(self):
        fields = Connector.schema_from_field_values({"A": 3, "B": "string", "C": False})
        assert fields == {"A": "INTEGER", "B": "TEXT", "C": "BOOLEAN"}

    def test_schema_from_value_datetime(self):
        datetime_type = Connector.field_value_to_type(np.datetime64("NaT"))
        assert Connector.type_mapping[datetime_type] == "TIMESTAMP"

    def test_expand_range_region_column_schema_datetime(self):
        result = Connector.expand_region_column_schema(["Range"], {"xdim": np.datetime64})
        assert result == {"start_xdim": "TIMESTAMP", "end_xdim": "TIMESTAMP"}

    def test_expand_range_region_column_schema_datetimes(self):
        result = Connector.expand_region_column_schema(
            ["Range"], {"xdim": np.datetime64, "ydim": int}
        )
        expected = {
            "start_xdim": "TIMESTAMP",
            "end_xdim": "TIMESTAMP",
            "start_ydim": "INTEGER",
            "end_ydim": "INTEGER",
        }
        assert result == expected

    def test_generate_schema(self):
        region_types = [["Range"]]
        kdim_dtypes = [{"xdim": np.datetime64, "ydim": int}]
        result = Connector.generate_schema(
            AutoIncrementKey(), region_types, kdim_dtypes, {"description": str}
        )
        expected = {
            "id": "INTEGER PRIMARY KEY AUTOINCREMENT",
            "start_xdim": "TIMESTAMP",
            "start_ydim": "INTEGER",
            "end_xdim": "TIMESTAMP",
            "end_ydim": "INTEGER",
            "description": "TEXT",
        }
        assert result == expected


class TestSQLiteDB:
    def test_setup(self, database):
        assert database.con is not None

    def test_initialized(self, database):
        assert not database.uninitialized

    def test_add_row(self, database, request):
        id1 = database.primary_key(database)
        start = pd.Timestamp("2022-06-01")
        end = pd.Timestamp("2022-06-03")
        description = "A description"
        insertion = {"uuid": id1, "description": description, "start": start, "end": end}
        df = pd.DataFrame(
            {
                "uuid": [database.primary_key.cast(id1)],
                "description": [description],
                "start": [start],
                "end": [end],
            }
        ).set_index("uuid")
        database.add_row(**insertion)
        pd.testing.assert_frame_equal(database.load_dataframe(), df)

    def test_add_three_rows_delete_one(self, database):
        id1 = database.primary_key(database)
        insertion1 = {
            "uuid": id1,
            "description": "A description",
            "start": pd.Timestamp("2022-06-01"),
            "end": pd.Timestamp("2022-06-03"),
        }

        id2 = database.primary_key(database, [id1])
        insertion2 = {
            "uuid": id2,
            "description": "A 2nd description",
            "start": pd.Timestamp("2024-06-01"),
            "end": pd.Timestamp("2024-06-03"),
        }

        id3 = database.primary_key(database, [id2])
        insertion3 = {
            "uuid": id3,
            "description": "A 3rd description",
            "start": pd.Timestamp("2026-06-01"),
            "end": pd.Timestamp("2026-06-03"),
        }

        df_data = {
            "uuid": map(database.primary_key.cast, [insertion1["uuid"], insertion3["uuid"]]),
            "description": [insertion1["description"], insertion3["description"]],
            "start": [insertion1["start"], insertion3["start"]],
            "end": [insertion1["end"], insertion3["end"]],
        }
        df = pd.DataFrame(df_data).set_index("uuid")
        database.add_row(**insertion1)
        database.add_row(**insertion2)
        database.add_row(**insertion3)
        database.delete_row(id2)
        pd.testing.assert_frame_equal(database.load_dataframe(), df)



from __future__ import annotations

import datetime as dt

import numpy as np
import pandas as pd
import pytest

from holonote.annotate import Annotator


class TestBasicRange1DAnnotator:
    @pytest.mark.skip("Need to add validation to set_regions")
    def test_point_insertion_exception(self, annotator_range1d):
        timestamp = np.datetime64("2022-06-06")
        expected_msg = "Only 'point' region allowed for 'set_regions'"
        with pytest.raises(ValueError, match=expected_msg):
            annotator_range1d.set_regions(TIME=timestamp)

    def test_insertion_edit_table_columns(self, annotator_range1d):
        annotator_range1d.set_regions(
            TIME=(np.datetime64("2022-06-06"), np.datetime64("2022-06-08"))
        )
        annotator_range1d.add_annotation(description="A test annotation!")
        commits = annotator_range1d.commit(return_commits=True)
        assert len(commits) == 1, "Only one insertion commit made "
        annotator_range1d.commit(return_commits=True)
        assert commits[0]["operation"] == "insert"
        assert set(commits[0]["kwargs"].keys()) == set(annotator_range1d.connector.columns)

    def test_range_insertion_values(self, annotator_range1d) -> None:
        start, end = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
        annotator_range1d.set_regions(TIME=(start, end))
        annotator_range1d.add_annotation(description="A test annotation!")
        commits = annotator_range1d.commit(return_commits=True)
        assert len(commits) == 1, "Only one insertion commit made"
        kwargs = commits[0]["kwargs"]
        assert "uuid" in kwargs, "Expected uuid primary key in kwargs"
        kwargs.pop("uuid")
        assert kwargs, {"description": "A test annotation!", "start_TIME": start, "end_TIME": end}

    def test_range_commit_insertion(self, annotator_range1d):
        start, end = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
        description = "A test annotation!"
        annotator_range1d.set_regions(TIME=(start, end))
        annotator_range1d.add_annotation(description=description)
        annotator_range1d.commit(return_commits=True)

        df = pd.DataFrame(
            {
                "uuid": pd.Series(annotator_range1d.df.index[0], dtype=object),
                "start_TIME": [start],
                "end_TIME": [end],
                "description": [description],
            }
        ).set_index("uuid")

        sql_df = annotator_range1d.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)

    def test_range_addition_deletion_by_uuid(self, annotator_range1d):
        start1, end1 = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
        start2, end2 = np.datetime64("2023-06-06"), np.datetime64("2023-06-08")
        start3, end3 = np.datetime64("2024-06-06"), np.datetime64("2024-06-08")
        annotator_range1d.set_regions(TIME=(start1, end1))
        annotator_range1d.add_annotation(description="Annotation 1")
        annotator_range1d.set_regions(TIME=(start2, end2))
        annotator_range1d.add_annotation(description="Annotation 2", uuid="08286429")
        annotator_range1d.set_regions(TIME=(start3, end3))
        annotator_range1d.add_annotation(description="Annotation 3")
        annotator_range1d.commit(return_commits=True)
        sql_df = annotator_range1d.connector.load_dataframe()
        assert set(sql_df["description"]) == {"Annotation 1", "Annotation 2", "Annotation 3"}
        deletion_index = sql_df.index[1]
        annotator_range1d.delete_annotation(deletion_index)
        annotator_range1d.commit(return_commits=True)
        sql_df = annotator_range1d.connector.load_dataframe()
        assert set(sql_df["description"]) == {"Annotation 1", "Annotation 3"}


class TestBasicRange2DAnnotator:
    @pytest.mark.skip("Need to add validation to set_regions")
    def test_point_insertion_exception(self, annotator_range2d):
        x, y = 0.5, 0.5
        expected_msg = "Only 'point' region allowed for 'set_regions'"
        with pytest.raises(ValueError, match=expected_msg):
            annotator_range2d.set_regions(x=x, y=y)

    def test_insertion_edit_table_columns(self, annotator_range2d):
        annotator_range2d.set_regions(x=(-0.25, 0.25), y=(-0.1, 0.1))
        annotator_range2d.add_annotation(description="A test annotation!")
        commits = annotator_range2d.commit(return_commits=True)
        assert len(commits) == 1, "Only one insertion commit made "
        assert commits[0]["operation"] == "insert"
        assert set(commits[0]["kwargs"].keys()) == set(annotator_range2d.connector.columns)

    def test_range_insertion_values(self, annotator_range2d):
        startx, endx, starty, endy = -0.25, 0.25, -0.1, 0.1
        annotator_range2d.set_regions(x=(startx, endx), y=(starty, endy))
        annotator_range2d.add_annotation(description="A test annotation!")
        commits = annotator_range2d.commit(return_commits=True)
        assert len(commits) == 1, "Only one insertion commit made"
        kwargs = commits[0]["kwargs"]
        assert "uuid" in kwargs, "Expected uuid primary key in kwargs"
        kwargs.pop("uuid")
        assert kwargs == {
            "description": "A test annotation!",
            "start_x": startx,
            "end_x": endx,
            "start_y": starty,
            "end_y": endy,
        }

    def test_range_commit_insertion(self, annotator_range2d):
        startx, endx, starty, endy = -0.25, 0.25, -0.1, 0.1
        description = "A test annotation!"
        annotator_range2d.set_regions(x=(startx, endx), y=(starty, endy))
        annotator_range2d.add_annotation(description=description)
        annotator_range2d.commit(return_commits=True)

        df = pd.DataFrame(
            {
                "uuid": pd.Series(annotator_range2d.df.index[0], dtype=object),
                "start_x": [startx],
                "start_y": [starty],
                "end_x": [endx],
                "end_y": [endy],
                "description": [description],
            }
        ).set_index("uuid")

        sql_df = annotator_range2d.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)

    def test_range_addition_deletion_by_uuid(self, annotator_range2d):
        startx1, endx1, starty1, endy1 = -0.251, 0.251, -0.11, 0.11
        startx2, endx2, starty2, endy2 = -0.252, 0.252, -0.12, 0.12
        startx3, endx3, starty3, endy3 = -0.253, 0.253, -0.13, 0.13
        annotator_range2d.set_regions(x=(startx1, endx1), y=(starty1, endy1))
        annotator_range2d.add_annotation(description="Annotation 1")
        annotator_range2d.set_regions(x=(startx2, endx2), y=(starty2, endy2))
        annotator_range2d.add_annotation(description="Annotation 2", uuid="08286429")
        annotator_range2d.set_regions(x=(startx3, endx3), y=(starty3, endy3))
        annotator_range2d.add_annotation(description="Annotation 3")
        annotator_range2d.commit(return_commits=True)
        sql_df = annotator_range2d.connector.load_dataframe()
        assert set(sql_df["description"]) == {"Annotation 1", "Annotation 2", "Annotation 3"}
        deletion_index = sql_df.index[1]
        annotator_range2d.delete_annotation(deletion_index)
        annotator_range2d.commit(return_commits=True)
        sql_df = annotator_range2d.connector.load_dataframe()
        assert set(sql_df["description"]) == {"Annotation 1", "Annotation 3"}


class TestBasicPoint1DAnnotator:
    def test_insertion_edit_table_columns(self, annotator_point1d):
        annotator_point1d.set_regions(TIME=np.datetime64("2022-06-06"))
        annotator_point1d.add_annotation(description="A test annotation!")
        commits = annotator_point1d.commit(return_commits=True)
        assert len(commits) == 1, "Only one insertion commit made "
        annotator_point1d.commit(return_commits=True)
        assert commits[0]["operation"] == "insert"
        assert set(commits[0]["kwargs"].keys()) == set(annotator_point1d.connector.columns)

    @pytest.mark.skip("Need to add validation to set_regions")
    def test_range_insertion_exception(self, annotator_point1d):
        start, end = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
        msg = "Only 'range' region allowed for 'set_range'"
        with pytest.raises(ValueError, match=msg):
            annotator_point1d.set_regions(TIME=(start, end))

    def test_point_insertion_values(self, annotator_point1d):
        timestamp = np.datetime64("2022-06-06")
        annotator_point1d.set_regions(TIME=timestamp)
        annotator_point1d.add_annotation(description="A test annotation!")
        commits = annotator_point1d.commit(return_commits=True)
        assert len(commits) == 1, "Only one insertion commit made"
        kwargs = commits[0]["kwargs"]
        assert "uuid" in kwargs, "Expected uuid primary key in kwargs"
        kwargs.pop("uuid")
        assert kwargs == {"description": "A test annotation!", "point_TIME": timestamp}

    def test_point_commit_insertion(self, annotator_point1d):
        timestamp = np.datetime64("2022-06-06")
        description = "A test annotation!"
        annotator_point1d.set_regions(TIME=timestamp)
        annotator_point1d.add_annotation(description=description)
        annotator_point1d.commit(return_commits=True)

        df = pd.DataFrame(
            {
                "uuid": pd.Series(annotator_point1d.df.index[0], dtype=object),
                "point_TIME": [timestamp],
                "description": [description],
            }
        ).set_index("uuid")

        sql_df = annotator_point1d.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)

    def test_point_addition_deletion_by_uuid(self, annotator_point1d):
        ts1 = np.datetime64("2022-06-06")
        ts2 = np.datetime64("2023-06-06")
        ts3 = np.datetime64("2024-06-06")
        annotator_point1d.set_regions(TIME=ts1)
        annotator_point1d.add_annotation(description="Annotation 1")
        annotator_point1d.set_regions(TIME=ts2)
        annotator_point1d.add_annotation(description="Annotation 2", uuid="08286429")
        annotator_point1d.set_regions(TIME=ts3)
        annotator_point1d.add_annotation(description="Annotation 3")
        annotator_point1d.commit(return_commits=True)
        sql_df = annotator_point1d.connector.load_dataframe()
        assert set(sql_df["description"]) == {"Annotation 1", "Annotation 2", "Annotation 3"}
        deletion_index = sql_df.index[1]
        annotator_point1d.delete_annotation(deletion_index)
        annotator_point1d.commit(return_commits=True)
        sql_df = annotator_point1d.connector.load_dataframe()
        assert set(sql_df["description"]) == {"Annotation 1", "Annotation 3"}


class TestBasicPoint2DAnnotator:
    def test_insertion_edit_table_columns(self, annotator_point2d):
        annotator_point2d.set_regions(x=-0.25, y=0.1)
        annotator_point2d.add_annotation(description="A test annotation!")
        commits = annotator_point2d.commit(return_commits=True)
        assert len(commits) == 1, "Only one insertion commit made "
        assert commits[0]["operation"] == "insert"
        assert set(commits[0]["kwargs"].keys()) == set(annotator_point2d.connector.columns)

    @pytest.mark.skip("Need to add validation to set_regions")
    def test_range_insertion_exception(self, annotator_point2d):
        x1, x2, y1, y2 = -0.25, 0.25, -0.3, 0.3
        expected_msg = "Only 'range' region allowed for 'set_range'"
        with pytest.raises(ValueError, match=expected_msg):
            annotator_point2d.set_regions(x=(x1, x2), y=(y1, y2))

    def test_point_insertion_values(self, annotator_point2d):
        x, y = 0.5, 0.3
        annotator_point2d.set_regions(x=x, y=y)
        annotator_point2d.add_annotation(description="A test annotation!")
        commits = annotator_point2d.commit(return_commits=True)
        assert len(commits) == 1, "Only one insertion commit made"
        kwargs = commits[0]["kwargs"]
        assert "uuid" in kwargs, "Expected uuid primary key in kwargs"
        kwargs.pop("uuid")
        assert kwargs == {"description": "A test annotation!", "point_x": x, "point_y": y}

    def test_point_commit_insertion(self, annotator_point2d):
        x, y = 0.5, 0.3
        description = "A test annotation!"
        annotator_point2d.set_regions(x=x, y=y)
        annotator_point2d.add_annotation(description=description)
        annotator_point2d.commit(return_commits=True)

        df = pd.DataFrame(
            {
                "uuid": pd.Series(annotator_point2d.df.index[0], dtype=object),
                "point_x": [x],
                "point_y": [y],
                "description": [description],
            }
        ).set_index("uuid")

        sql_df = annotator_point2d.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)

    def test_point_addition_deletion_by_uuid(self, annotator_point2d):
        x1, y1 = 0.2, 0.2
        x2, y2 = 0.3, 0.3
        x3, y3 = 0.4, 0.4
        annotator_point2d.set_regions(x=x1, y=y1)
        annotator_point2d.add_annotation(description="Annotation 1")
        annotator_point2d.set_regions(x=x2, y=y2)
        annotator_point2d.add_annotation(description="Annotation 2", uuid="08286429")
        annotator_point2d.set_regions(x=x3, y=y3)
        annotator_point2d.add_annotation(description="Annotation 3")
        annotator_point2d.commit(return_commits=True)
        sql_df = annotator_point2d.connector.load_dataframe()
        assert set(sql_df["description"]) == {"Annotation 1", "Annotation 2", "Annotation 3"}
        deletion_index = sql_df.index[1]
        annotator_point2d.delete_annotation(deletion_index)
        annotator_point2d.commit(return_commits=True)
        sql_df = annotator_point2d.connector.load_dataframe()
        assert set(sql_df["description"]) == {"Annotation 1", "Annotation 3"}


@pytest.mark.parametrize("fields", [["test"], ["test1", "test2"]])
def test_connector_use_annotator_fields(conn_sqlite_uuid, fields):
    annotator = Annotator({"TIME": float}, connector=conn_sqlite_uuid, fields=fields)

    assert annotator.fields == fields
    assert annotator.connector.fields == fields


def test_connector_use_annotator_fields_default(conn_sqlite_uuid):
    annotator = Annotator({"TIME": float}, connector=conn_sqlite_uuid)

    assert annotator.fields == ["description"]
    assert annotator.connector.fields == ["description"]


def test_fields_with_spaces(conn_sqlite_uuid):
    fields = ["A Field with space"]
    annotator = Annotator({"A": float, "B": float}, fields=fields, connector=conn_sqlite_uuid)
    assert annotator.annotation_table._field_df.columns == fields


def test_static_fields(conn_sqlite_uuid):
    static_fields = {"static": "1"}
    annotator = Annotator({"TIME": float}, static_fields=static_fields, connector=conn_sqlite_uuid)

    assert annotator.static_fields == static_fields
    assert "static" in annotator.annotation_table._field_df.columns
    assert "static" in annotator.df
    assert annotator.connector.fields == ["description", "static"]

    # Create new entry
    start, end = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
    annotator.set_regions(TIME=(start, end))
    annotator.add_annotation(description="test")
    output = annotator.df
    assert output["static"].iloc[0] == "1"
    assert output["description"].iloc[0] == "test"
    assert output.shape[0] == 1

    # Update entry
    annotator.update_annotation_fields(output.index[0], description="test2")
    output = annotator.df
    assert output["static"].iloc[0] == "1"
    assert output["description"].iloc[0] == "test2"
    assert output.shape[0] == 1


@pytest.mark.parametrize("dtype", [dt.datetime, dt.date])
def test_spec_datetime_date(conn_sqlite_uuid, dtype):
    annotator = Annotator(spec={"time": dtype}, connector=conn_sqlite_uuid)

    output = annotator.df.dtypes
    assert output["start[time]"] == np.dtype("datetime64[ns]")
    assert output["end[time]"] == np.dtype("datetime64[ns]")



from __future__ import annotations

from typing import Callable, Literal, TypedDict


class SpecItem(TypedDict):
    """Contains the type and region specification for a region.

    The type is a callable that can be used to convert the data to the
    correct type. This could be `np.datetime64` or `float` for example.

    The region specification is either "range", "point", or "geometry".
    """

    type: Callable
    region: Literal["range", "point", "geometry"]


SpecDict = dict[str, SpecItem]



from __future__ import annotations

import datetime as dt
import sqlite3
import uuid
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import param

if TYPE_CHECKING:
    from .typing import SpecDict


class PrimaryKey(param.Parameterized):
    """
    Generator of the primary key used to keep track of annotations in
    HoloViews.

    The generated key is used to reference annotations until they are
    committed, at which point they may 1) be inserted in the database as
    the primary key value (policy='insert') 2) are checked against the
    primary key value chosen by the database which is expected to match
    in most cases.

    In real situations where the key is chosen by the database, the key
    generated will *not* always match the actual key assigned. The
    policy parameter decides the resulting behavior in these cases.
    """

    field_name = param.String(default="id", allow_None=False)

    policy = param.ObjectSelector(
        default="ignore-mismatch",
        objects=["insert", "ignore-mismatch", "warn-mismatch", "error-mismatch"],
    )

    schema = param.String(default="INTEGER PRIMARY KEY", constant=True, allow_None=False)

    connector_class = param.String(default="SQLiteDB")

    def __call__(self, connector, key_list=None):
        """
        The key list is the current list of index values that are
        outstanding (i.e. have not been committed).
        """
        raise NotImplementedError

    def validate(self, database_key, local_key):
        if self.policy in ["insert", "ignore-mismatch"]:
            return

        if database_key != local_key:
            print("MISMATCH")

    def cast(self, value):
        "Cast a user supplied value to a known supported type"
        raise NotImplementedError


class AutoIncrementKey(PrimaryKey):
    """
    AUTOINCREMENT needed to prevent reuse of ids from deleted rows:
    https://www.sqlite.org/autoinc.html
    """

    field_name = param.String(default="id", allow_None=False)

    policy = param.ObjectSelector(
        default="ignore-mismatch",
        objects=["insert", "ignore-mismatch", "warn-mismatch", "error-mismatch"],
    )

    schema = param.String(
        default="INTEGER PRIMARY KEY AUTOINCREMENT", constant=True, allow_None=False
    )

    def __call__(self, connector, key_list=None):
        key_list_max = max(key_list) if key_list else 0
        connector_max = connector.max_rowid()
        connector_max = 0 if connector_max is None else connector_max
        max_rowid = max(key_list_max, connector_max)
        return max_rowid + 1

    def cast(self, value):
        return int(value)  # e.g. np.int64 from a pandas index won't work!


class UUIDHexStringKey(PrimaryKey):  # Probably the better default
    """
    Example of 'insert' policy where the generated primary key value can
    be inserted in the database as uuids are independent and not
    expected to clash.
    """

    field_name = param.String(default="uuid", allow_None=False)

    policy = param.ObjectSelector(
        default="insert", objects=["insert", "ignore-mismatch", "warn-mismatch", "error-mismatch"]
    )

    schema = param.String("TEXT PRIMARY KEY", constant=True, allow_None=False)

    length = param.Integer(default=32, bounds=(4, 32))

    def __call__(self, connector, key_list=None):
        return uuid.uuid4().hex[: self.length]

    def cast(self, value):
        return str(value)


class UUIDBinaryKey(PrimaryKey):
    """
    Example of 'insert' policy where the generated primary key value can
    be inserted in the database as uuids are independent and not
    expected to clash.
    """

    field_name = param.String(default="uuid", allow_None=False)

    policy = param.ObjectSelector(
        default="insert", objects=["insert", "ignore-mismatch", "warn-mismatch", "error-mismatch"]
    )

    schema = param.String("BINARY PRIMARY KEY", constant=True, allow_None=False)

    def __call__(self, connector, key_list=None):
        return uuid.uuid4().bytes

    def cast(self, value):
        return bytes(value)


class WidgetKey(PrimaryKey):
    """
    Placeholder for a concept where the user can insert a primary key
    value via a widget.
    """


class Connector(param.Parameterized):
    """
    Base class that support the auto-generated default schema
    """

    primary_key = param.Parameter(default=UUIDHexStringKey(), allow_None=False)

    commit_hook = param.Parameter(default=None, doc="Callback, applies default schema if None")

    fields = param.List(default=None, doc="List of column names for domain-specific fields")

    transforms = param.Dict(
        default={
            "insert": lambda x: x,
            "update": lambda x: x,
            "delete": lambda x: x,
            "save": lambda x: x,
            "load": lambda x: x,
        }
    )

    operation_mapping = {}  # Mapping from operation type to corresponding connector method

    type_mapping = {
        bool: "BOOLEAN",
        str: "TEXT",
        float: "REAL",
        int: "INTEGER",
        np.datetime64: "TIMESTAMP",
        dt.date: "TIMESTAMP",
        dt.datetime: "TIMESTAMP",
        param.Integer: "INTEGER",
        param.Number: "REAL",
        param.String: "TEXT",
        param.Boolean: "BOOLEAN",
        np.dtype("datetime64[ns]"): "TIMESTAMP",
        np.dtype("<M8"): "TIMESTAMP",
        np.float64: "REAL",
        np.int64: "INTEGER",
    }

    @classmethod
    def field_value_to_type(cls, value):
        if isinstance(value, list):
            assert all(isinstance(el, str) for el in value), "Only string enums supported"
            return str
        elif hasattr(value, "dtype"):
            return value.dtype
        elif type(value) in cls.type_mapping:
            return type(value)
        elif isinstance(value, param.Parameter) and value.default is not None:
            return type(value.default)
        else:
            msg = f"Connector cannot handle type {type(value)!s}"
            raise TypeError(msg)

    @classmethod
    def schema_from_field_values(cls, fields):
        "Given a dictionary of fields fields and values return the field schemas"
        return {
            name: cls.type_mapping[cls.field_value_to_type(value)]
            for name, value in fields.items()
        }

    @classmethod
    def schema_from_field_types(cls, fields):
        "Given a dictionary of fields fields and values return the field schemas"
        return {name: cls.type_mapping[value] for name, value in fields.items()}

    @classmethod
    def expand_region_column_schema(cls, region_types, kdim_dtypes):
        region_type_column_prefixes = {
            "Range": ("start", "end"),
            "Point": ("point",),
            "point": ("point",),
            "range": ("start", "end"),
        }
        fields = {}
        for region_type in region_types:
            for column in region_type_column_prefixes[region_type]:
                for kdim, kdim_dtype in kdim_dtypes.items():
                    fields[f"{column}_{kdim}"] = cls.type_mapping[kdim_dtype]
        return fields

    @classmethod
    def generate_schema(cls, primary_key, all_region_types, all_kdim_dtypes, field_types):
        schemas = {primary_key.field_name: primary_key.schema}
        region_schemas = {}
        for region_types, kdim_dtypes in zip(all_region_types, all_kdim_dtypes):
            region_schemas = dict(
                region_schemas, **cls.expand_region_column_schema(region_types, kdim_dtypes)
            )
        schemas = dict(schemas, **region_schemas)
        return dict(schemas, **cls.schema_from_field_types(field_types))

    def _incompatible_schema_check(self, expected_keys, columns, fields, region_type):
        msg_prefix = (
            "Unable to read annotations that were stored with a "
            "schema inconsistent with the current settings:"
        )
        msg_suffix = f"Columns found: {columns}"

        missing_field_columns = set(fields) - set(columns)
        if missing_field_columns:
            msg = f"{msg_prefix} Missing field columns {missing_field_columns}. {msg_suffix}"
            raise Exception(msg)

        non_field_columns = set(columns) - set(fields)
        missing_region_columns = set(expected_keys) - non_field_columns
        if missing_region_columns:
            msg = f"{msg_prefix} Missing {region_type!r} region columns {missing_region_columns}. {msg_suffix}"
            raise Exception(msg)

    def _create_column_schema(self, spec: SpecDict, fields: list[str]) -> None:
        field_dtypes = {col: str for col in fields}  # FIXME - generalize
        all_region_types = [{v["region"] for v in spec.values()}]
        all_kdim_dtypes = [{k: v["type"] for k, v in spec.items()}]
        schema = self.generate_schema(
            self.primary_key, all_region_types, all_kdim_dtypes, field_dtypes
        )
        self.column_schema = schema


class SQLiteDB(Connector):
    """
    Simple example of a Connector without dependencies, using sqlite3.

    column_schema is a dictionary from column name to SQL CREATE TABLE type declaration.
    """

    filename = param.String(default="annotations.db")

    table_name = param.String(default="annotations")

    column_schema = param.Dict(default={})

    operation_mapping = {
        "insert": "add_row",
        "delete": "delete_row",
        "save": "add_rows",
        "update": "update_row",
    }

    def __init__(self, column_schema=None, connect=True, **params):
        """
        First key in column_schema is assumed to the primary key field if not explicitly specified.
        """
        if column_schema is None:
            column_schema = {}

        params["column_schema"] = column_schema
        self.con, self.cursor = None, None
        super().__init__(**params)

        if connect:
            self._initialize(column_schema, create_table=False)

    def _initialize(self, column_schema, create_table=True):
        if self.con is None:
            self.con = sqlite3.connect(
                self.filename, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES
            )
            self.cursor = self.con.cursor()  # should be context manager
        if create_table:
            self.create_table(column_schema=column_schema)

    @property
    def uninitialized(self):
        if self.con is not None:
            return self.table_name not in self.get_tables()
        return True

    @property
    def columns(self):
        "Return names of columns"
        result = self.cursor.execute("PRAGMA table_info('%s')" % self.table_name).fetchall()
        return list(zip(*result))[1]

    def max_rowid(self):
        return self.cursor.execute(f"SELECT max(ROWID) from {self.table_name}").fetchone()[0]

    def initialize(self, column_schema):
        self.column_schema = column_schema
        self._initialize(column_schema)

    def load_dataframe(self):
        uninitialized = self.cursor is None
        if uninitialized:
            self._initialize({}, create_table=False)

        raw_df = pd.read_sql_query(f"SELECT * FROM {self.table_name}", self.con)
        # dtype={self.primary_key.field_name:self.primary_key.dtype})
        df = raw_df.set_index(self.primary_key.field_name)
        if uninitialized:
            self.con, self.cursor = None, None
        return df

    def get_tables(self):
        res = self.cursor.execute("SELECT name FROM sqlite_master")
        return [el[0] for el in res.fetchmany()]

    def create_table(self, column_schema=None):
        column_schema = column_schema if column_schema else self.column_schema
        column_spec = ",\n".join([f'"{name}" {spec}' for name, spec in column_schema.items()])
        create_table_sql = f"CREATE TABLE IF NOT EXISTS {self.table_name} (" + column_spec + ");"
        self.cursor.execute(create_table_sql)
        self.con.commit()

    def delete_table(self):
        self.cursor.execute(f"DROP TABLE IF EXISTS {self.table_name}")
        self.con.commit()

    def add_rows(self, field_list):  # Used execute_many
        for field in field_list:
            self.add_row(**field)

    def add_row(self, **fields):
        # Note, missing fields will be set as NULL
        columns = self.columns
        field_values = [fields.get(col, None) for col in self.columns]
        field_values = [
            pd.to_datetime(el) if isinstance(el, np.datetime64) else el for el in field_values
        ]
        field_values = [
            el.to_pydatetime() if isinstance(el, pd.Timestamp) else el for el in field_values
        ]

        if self.primary_key.policy != "insert":
            field_values = field_values[1:]
            columns = columns[1:]

        placeholders = ", ".join(["?"] * len(field_values))
        self.cursor.execute(
            f"INSERT INTO {self.table_name} {columns!s} VALUES({placeholders});", field_values
        )
        self.primary_key.validate(self.cursor.lastrowid, fields[self.primary_key.field_name])
        self.con.commit()

    def delete_all_rows(self):
        "Obviously a destructive operation!"
        self.cursor.execute(f"DELETE FROM {self.table_name};")
        self.con.commit()

    def delete_row(self, id_val):
        self.cursor.execute(
            f"DELETE FROM {self.table_name} WHERE {self.primary_key.field_name} = ?",
            (self.primary_key.cast(id_val),),
        )
        self.con.commit()

    def update_row(self, **updates):  # updates as a dictionary OR remove posarg?
        assert self.primary_key.field_name in updates
        id_val = updates.pop(self.primary_key.field_name)
        set_updates = ", ".join('"' + k + '"' + " = ?" for k in updates)
        query = f'UPDATE {self.table_name} SET {set_updates} WHERE "{self.primary_key.field_name}" = ?;'
        self.cursor.execute(query, [*updates.values(), id_val])
        self.con.commit()



import inspect
import os
import warnings

import holoviews as hv
import param

__all__ = (
    "find_stack_level",
    "warn",
)


def warn(message, category=None, stacklevel=None):
    if stacklevel is None:
        stacklevel = find_stack_level()

    warnings.warn(message, category, stacklevel=stacklevel)


def find_stack_level():
    """
    Find the first place in the stack that is not inside Holoviews and Param.
    Inspired by: pandas.util._exceptions.find_stack_level
    """

    import holonote as hn

    pkg_dir = os.path.dirname(hn.__file__)
    test_dir = os.path.join(pkg_dir, "tests")
    hv_dir = os.path.dirname(hv.__file__)
    param_dir = os.path.dirname(param.__file__)

    try:
        frame = inspect.currentframe()
        stacklevel = 0
        while frame:
            fname = inspect.getfile(frame)
            if fname.startswith((pkg_dir, hv_dir, param_dir)) and not fname.startswith(test_dir):
                frame = frame.f_back
                stacklevel += 1
            else:
                break
    finally:
        del frame

    return stacklevel



from __future__ import annotations

from . import annotate, editor  # noqa: F401

# Define '__version__'
try:
    # If setuptools_scm is installed (e.g. in a development environment with
    # an editable install), then use it to determine the version dynamically.
    from setuptools_scm import get_version

    # This will fail with LookupError if the package is not installed in
    # editable mode or if Git is not installed.
    __version__ = get_version(root="..", relative_to=__file__)
except (ImportError, LookupError):
    # As a fallback, use the version that is hard-coded in the file.
    try:
        from ._version import __version__
    except ModuleNotFoundError:
        # The user is probably trying to run this without having installed
        # the package.
        __version__ = "0.0.0+unknown"



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Any

import holoviews as hv
import pandas as pd
import param
from bokeh.models.tools import BoxSelectTool, HoverTool, Tool

from .._warnings import warn
from .connector import Connector, SQLiteDB
from .table import AnnotationTable

if TYPE_CHECKING:
    from .typing import SpecDict


class Indicator:
    """
    Collection of class methods that express annotation data as final
    displayed (vectorized) HoloViews object.
    """

    range_style = {"color": "red", "alpha": 0.4, "apply_ranges": False}
    point_style = {"color": "red", "alpha": 0.4, "apply_ranges": False}
    indicator_highlight = {"alpha": (0.7, 0.2)}

    edit_range_style = {"alpha": 0.4, "line_alpha": 1, "line_width": 1, "line_color": "black"}
    edit_point_style = {"alpha": 0.4, "line_alpha": 1, "line_color": "black"}

    @classmethod
    def indicator_style(cls, range_style, point_style, highlighters):
        return (
            hv.opts.Rectangles(**dict(range_style, **highlighters)),
            hv.opts.VSpans(**dict(range_style, **highlighters)),
            hv.opts.HSpans(**dict(range_style, **highlighters)),
            hv.opts.VLines(**dict(range_style, **highlighters)),
            hv.opts.HLines(**dict(range_style, **highlighters)),
        )

    @classmethod
    def region_style(cls, edit_range_style, edit_point_style):
        return (
            hv.opts.Rectangles(**edit_range_style),
            hv.opts.VSpans(**edit_range_style),
            hv.opts.HSpans(**edit_range_style),
            hv.opts.VLines(**edit_range_style),
            hv.opts.HLines(**edit_range_style),
        )

    @classmethod
    def points_1d(cls, data, region_labels, fields_labels, invert_axes=False):
        "Vectorizes point regions to VLines. Note does not support hover info"
        vdims = [*fields_labels, data.index.name]
        element = hv.VLines(data.reset_index(), kdims=region_labels, vdims=vdims)
        hover = cls._build_hover_tool(data)
        return element.opts(tools=[hover])

    @classmethod
    def points_2d(cls, data, region_labels, fields_labels, invert_axes=False):
        "Vectorizes point regions to VLines * HLines. Note does not support hover info"
        msg = "2D point regions not supported yet"
        raise NotImplementedError(msg)
        vdims = [*fields_labels, data.index.name]
        element = hv.Points(data.reset_index(), kdims=region_labels, vdims=vdims)
        hover = cls._build_hover_tool(data)
        return element.opts(tools=[hover])

    @classmethod
    def ranges_2d(cls, data, region_labels, fields_labels, invert_axes=False):
        "Vectorizes an nd-overlay of range_2d rectangles."
        kdims = [region_labels[i] for i in (0, 2, 1, 3)]  # LBRT format
        vdims = [*fields_labels, data.index.name]
        element = hv.Rectangles(data.reset_index(), kdims=kdims, vdims=vdims)
        cds_map = dict(zip(region_labels, ("left", "right", "bottom", "top")))
        hover = cls._build_hover_tool(data, cds_map)
        return element.opts(tools=[hover])

    @classmethod
    def ranges_1d(cls, data, region_labels, fields_labels, invert_axes=False):
        """
        Vectorizes an nd-overlay of range_1d rectangles.

        NOTE: Should use VSpans once available!
        """
        vdims = [*fields_labels, data.index.name]
        element = hv.VSpans(data.reset_index(), kdims=region_labels, vdims=vdims)
        hover = cls._build_hover_tool(data)
        return element.opts(tools=[hover])

    @classmethod
    def _build_hover_tool(self, data, cds_map=None) -> HoverTool:
        if cds_map is None:
            cds_map = {}
        tooltips, formatters = [], {}
        for dim in data.columns:
            cds_name = cds_map.get(dim, dim)
            if data[dim].dtype.kind == "M":
                tooltips.append((dim, f"@{{{cds_name}}}{{%F}}"))
                formatters[f"@{{{cds_name}}}"] = "datetime"
            else:
                tooltips.append((dim, f"@{{{cds_name}}}"))
        return HoverTool(tooltips=tooltips, formatters=formatters)


class AnnotatorInterface(param.Parameterized):
    """
    Baseclass that expresses the Python interface of an Annotator
    without using any holoviews components and without requiring
    display. Most of this API centers around how the Annotator interacts
    with the AnnotationTable.
    """

    selected_indices = param.List(default=[], doc="Indices of selected annotations")

    spec = param.Dict(default={}, doc="Specification of annotation types")

    fields = param.List(default=["description"], doc="List of fields", constant=True)

    static_fields = param.Dict(
        default={},
        constant=True,
        doc="Dictionary with key and value which will be added to each commit",
    )

    connector = param.ClassSelector(class_=Connector, allow_None=False)

    connector_class = SQLiteDB

    def __init__(self, spec, **params):
        if "connector" not in params:
            params["connector"] = self.connector_class()

        spec = self.normalize_spec(spec)

        super().__init__(spec=spec, **params)
        if set(self.fields) & set(self.static_fields):
            msg = "The values of fields and static_fields must not overlap"
            raise ValueError(msg)
        if self.connector.fields is None:
            self.connector.fields = self.all_fields
        self._region = {}
        self._last_region = None

        self.annotation_table = AnnotationTable()
        self.connector._create_column_schema(self.spec, self.all_fields)
        self.connector._initialize(self.connector.column_schema)
        self.annotation_table.load(self.connector, fields=self.connector.fields, spec=self.spec)

    @property
    def all_fields(self) -> list:
        """Return a list of all fields including static fields"""
        return [*self.fields, *self.static_fields]

    @classmethod
    def normalize_spec(self, input_spec: dict[str, Any]) -> SpecDict:
        """Normalize the spec to conform to SpecDict format

        Accepted input spec formats:
        spec = {
            # Range (two values)
            "A1": (np.float64, "range"),
            "A2": {"type": np.float64, "region": "range"},
            "A3": np.float64,  # Special case
            # Point
            "B1": (np.float64, "point"),
            "B2": {"type": np.float64, "region": "point"},
            # Geometry
            ("C1", "D1"): {"type": np.float64, "region": "geometry"},
            ("C2", "D2"): (np.float64, "geometry"),
        }
        """

        new_spec: SpecDict = {}
        for k, v in input_spec.items():
            if isinstance(v, dict):
                pass
            elif isinstance(v, tuple):
                v = {"type": v[0], "region": v[1]}
            else:
                v = {"type": v, "region": "range"}

            if v["region"] not in ["range", "point", "geometry"]:
                msg = "Region type must be range, point, or geometry."
                raise ValueError(msg)
            if v["region"] == "geometry" and not isinstance(k, tuple):
                msg = "Geometry region dimension must be a tuple."
                raise ValueError(msg)
            new_spec[k] = v

        return new_spec

    @property
    def df(self) -> pd.DataFrame:
        return self.annotation_table.get_dataframe(spec=self.spec)

    def get_dataframe(self, dims) -> pd.DataFrame:
        return self.annotation_table.get_dataframe(spec=self.spec, dims=dims)

    def refresh(self, clear=False):
        "Method to update display state of the annotator and optionally clear stale visual state"

    # Selecting annotations

    def select_by_index(self, *inds):
        "Set the selection state by the indices i.e. primary key values"
        self.selected_indices = list(inds)

    @property
    def selected_index(self):
        "Convenience property returning a single selected index (the first one) or None"
        return self.selected_indices[0] if len(self.selected_indices) > 0 else None

    @property
    def region(self):
        return self._region

    def set_regions(self, **items):
        self._set_regions(**items)

    def _set_regions(self, **items):
        """Updating regions"""
        # TODO: Validate values based on spec
        for dim, values in items.items():
            if dim not in self.spec:
                msg = f"Dimension {dim} not in spec"
                raise ValueError(msg)
            self._region[dim] = values

    def clear_regions(self):
        self._region = {}
        self._last_region = {}

    def _add_annotation(self, **fields):
        # Primary key specification is optional
        if self.connector.primary_key.field_name not in fields:
            index_val = self.connector.primary_key(
                self.connector, list(self.annotation_table._field_df.index)
            )
            fields[self.connector.primary_key.field_name] = index_val

        # Don't do anything if self.region is an empty dict
        if self.region and self.region != self._last_region:
            self.annotation_table.add_annotation(
                self._region, spec=self.spec, **fields, **self.static_fields
            )
            self._last_region = self._region.copy()

    def add_annotation(self, **fields):
        self._add_annotation(**fields)

    def update_annotation_region(self, index):
        self.annotation_table.update_annotation_region(self._region, index)

    def update_annotation_fields(self, index, **fields):
        self.annotation_table.update_annotation_fields(index, **fields, **self.static_fields)

    def delete_annotation(self, index):
        try:
            self.annotation_table.delete_annotation(index)
        except KeyError:
            msg = f"Annotation with index {index!r} does not exist."
            raise ValueError(msg) from None

        if index in self.selected_indices:
            self.selected_indices.remove(index)

    # Defining initial annotations

    def define_annotations(self, data: pd.DataFrame, **kwargs) -> None:
        # Will both set regions and add annotations. Can accept multiple inputs
        # if index is none a new index will be set.
        # if nothing is given it infer the regions and fields from the column header.
        # annotator.define_annotations(
        #     df, TIME=("start", "end"), description="description", index=None
        # )

        # kwargs = dict(A=("start", "end"), description="description")
        # kwargs = dict(A=("start", "end"))
        # kwargs = dict(A=("start", "end"), index=True)

        index = kwargs.pop("index", False)
        field_data = set(self.fields) & set(data.columns)
        f_keys = (set(self.fields) & kwargs.keys()) | field_data
        r_keys = (kwargs.keys() - f_keys) | (set(self.spec) & set(data.columns))
        pk = self.connector.primary_key

        for k, v in kwargs.items():
            if k == v:
                continue
            if v in field_data:
                msg = (
                    f"Input {v!r} has overlapping name with a field or spec. "
                    "This can give weird behavior. Consider renaming the input."
                )
                warn(msg)
        # assert len(set(data.columns) - f_keys - r_keys) == 0

        # Vectorize the conversion?
        for r in data.itertuples(index=index):
            regions = {
                k: tuple([getattr(r, a) for a in kwargs[k]])
                if isinstance(kwargs.get(k, k), tuple)
                else getattr(r, kwargs.get(k, k))
                for k in r_keys
            }
            fields = {k: getattr(r, kwargs.get(k, k)) for k in f_keys}
            if index:
                fields[pk.field_name] = pk.cast(r.Index)

            self._set_regions(**regions)
            self._add_annotation(**fields)

    # Snapshotting and reverting
    @property
    def has_snapshot(self) -> bool:
        return self.annotation_table.has_snapshot

    def revert_to_snapshot(self) -> None:
        self.annotation_table.revert_to_snapshot()

    def snapshot(self) -> None:
        self.annotation_table.snapshot()

    def commit(self, return_commits=False):
        # self.annotation_table.initialize_table(self.connector)  # Only if not in params
        commits = self.annotation_table.commits(self.connector)
        if return_commits:
            return commits


class AnnotationDisplay(param.Parameterized):
    kdims = param.List(
        default=["x"], bounds=(1, 3), constant=True, doc="Dimensions of the element"
    )

    indicator = Indicator

    _count = param.Integer(default=0, precedence=-1)

    def __init__(self, annotator: Annotator, **params) -> None:
        super().__init__(**params)

        self._annotation_count_stream = hv.streams.Params(
            parameterized=self,
            parameters=["_count"],
            transient=True,
        )

        self._selection_info = {}

        self._selection_enabled = True
        self._editable_enabled = True
        self._selected_values = []
        self._selected_options = []

        transient = False
        self._edit_streams = [
            hv.streams.BoundsXY(transient=transient),
            hv.streams.SingleTap(transient=transient),
            hv.streams.Lasso(transient=transient),
        ]

        self.annotator = weakref.proxy(annotator)
        self._set_region_types()
        self._element = self._make_empty_element()

    def _set_region_types(self) -> None:
        self.region_types = "-".join([self.annotator.spec[k]["region"] for k in self.kdims])

    @property
    def element(self):
        return self.overlay()

    @property
    def edit_tools(self) -> list[Tool]:
        tools = []
        if self.region_types == "range":
            tools.append(BoxSelectTool(dimensions="width"))
        elif self.region_types == "range-range":
            tools.append(BoxSelectTool())
        elif self.region_types == "point":
            tools.append(BoxSelectTool(dimensions="width"))
        elif self.region_types == "point-point":
            tools.append("tap")
        return tools

    @classmethod
    def _infer_kdim_dtypes(cls, element):
        if not isinstance(element, hv.Element):
            msg = "Supplied object {element} is not a bare HoloViews Element"
            raise ValueError(msg)
        kdim_dtypes = {}
        for kdim in element.dimensions(selection="key"):
            kdim_dtypes[str(kdim)] = type(element.dimension_values(kdim)[0])
        return kdim_dtypes

    def clear_indicated_region(self):
        "Clear any region currently indicated on the plot by the editor"
        self._edit_streams[0].event(bounds=None)
        self._edit_streams[1].event(x=None, y=None)
        self._edit_streams[2].event(geometry=None)
        self.annotator.clear_regions()

    def _make_empty_element(self) -> hv.Curve | hv.Image:
        El = hv.Curve if len(self.kdims) == 1 else hv.Image
        return El([], kdims=self.kdims).opts(apply_ranges=False)

    @property
    def selection_element(self) -> hv.Element:
        if not hasattr(self, "_selection_element"):
            self._selection_element = self._make_empty_element()
        return self._selection_element

    @property
    def selection_enabled(self) -> bool:
        return self._selection_enabled

    @selection_enabled.setter
    def selection_enabled(self, enabled: bool) -> None:
        self._selection_enabled = enabled

    @property
    def editable_enabled(self) -> bool:
        return self._editable_enabled

    @editable_enabled.setter
    def editable_enabled(self, enabled: bool) -> None:
        self._editable_enabled = enabled
        if not enabled:
            self.clear_indicated_region()

    def _filter_stream_values(self, bounds, x, y, geometry):
        if not self._editable_enabled:
            return (None, None, None, None)
        if self.region_types == "point" and bounds:
            x = (bounds[0] + bounds[2]) / 2
            y = None
            bounds = (x, 0, x, 0)
        elif "range" not in self.region_types:
            bounds = None

        # If selection enabled, tap stream used for selection not for creating point regions
        # if ('point' in self.region_types and self.selection_enabled) or 'point' not in self.region_types:
        if "point" not in self.region_types:
            x, y = None, None

        return bounds, x, y, geometry

    def _make_selection_editor(self) -> hv.DynamicMap:
        def inner(bounds, x, y, geometry):
            bounds, x, y, geometry = self._filter_stream_values(bounds, x, y, geometry)

            info = self.selection_element._get_selection_expr_for_stream_value(
                bounds=bounds, x=x, y=y, geometry=geometry
            )
            (dim_expr, bbox, region_element) = info

            self._selection_info = {
                "dim_expr": dim_expr,
                "bbox": bbox,
                "x": x,
                "y": y,
                "geometry": geometry,
                "region_element": region_element,
            }

            if bbox is not None:
                # self.annotator.set_regions will give recursion error
                self.annotator._set_regions(**bbox)

            kdims = list(self.kdims)
            if self.region_types == "point" and x is not None:
                self.annotator._set_regions(**{kdims[0]: x})
            if None not in [x, y]:
                if len(kdims) == 1:
                    self.annotator._set_regions(**{kdims[0]: x})
                elif len(kdims) == 2:
                    self.annotator._set_regions(**{kdims[0]: x, kdims[1]: y})
                else:
                    msg = "Only 1d and 2d supported for Points"
                    raise ValueError(msg)

            return region_element

        return hv.DynamicMap(inner, streams=self._edit_streams)

    def region_editor(self) -> hv.DynamicMap:
        if not hasattr(self, "_region_editor"):
            self._region_editor = self._make_selection_editor()
        return self._region_editor

    def _get_range_indices_by_position(self, **inputs) -> list[Any]:
        df = self.static_indicators.data
        if df.empty:
            return []

        # Because we reset_index in Indicators
        id_col = df.columns[0]

        for i, (k, v) in enumerate(inputs.items()):
            mask = (df[f"start[{k}]"] <= v) & (v < df[f"end[{k}]"])
            if i == 0:
                ids = set(df[mask][id_col])
            else:
                ids &= set(df[mask][id_col])
        return list(ids)

    def _get_point_indices_by_position(self, **inputs) -> list[Any]:
        """
        Simple algorithm for finding the closest point
        annotation to the given position.
        """

        df = self.static_indicators.data
        if df.empty:
            return []

        # Because we reset_index in Indicators
        id_col = df.columns[0]

        for i, (k, v) in enumerate(inputs.items()):
            nearest = (df[f"point[{k}]"] - v).abs().argmin()
            if i == 0:
                ids = {df.iloc[nearest][id_col]}
            else:
                ids &= {df.iloc[nearest][id_col]}
        return list(ids)

    def get_indices_by_position(self, **inputs) -> list[Any]:
        "Return primary key values matching given position in data space"
        if "range" in self.region_types:
            return self._get_range_indices_by_position(**inputs)
        elif "point" in self.region_types:
            return self._get_point_indices_by_position(**inputs)
        else:
            msg = f"{self.region_types} not implemented"
            raise NotImplementedError(msg)

    def register_tap_selector(self, element: hv.Element) -> hv.Element:
        def tap_selector(x, y) -> None:  # Tap tool must be enabled on the element
            # Only select the first
            inputs = {str(k): v for k, v in zip(self.kdims, (x, y))}
            indices = self.get_indices_by_position(**inputs)
            if indices:
                self.annotator.select_by_index(indices[0])
            else:
                self.annotator.select_by_index()

        tap_stream = hv.streams.Tap(source=element, transient=True)
        tap_stream.add_subscriber(tap_selector)
        return element

    def register_double_tap_clear(self, element: hv.Element) -> hv.Element:
        def double_tap_clear(x, y):
            self.clear_indicated_region()

        double_tap_stream = hv.streams.DoubleTap(source=element, transient=True)
        double_tap_stream.add_subscriber(double_tap_clear)
        return element

    def indicators(self) -> hv.DynamicMap:
        self.register_tap_selector(self._element)
        self.register_double_tap_clear(self._element)

        def inner(_count):
            return self.static_indicators

        return hv.DynamicMap(inner, streams=[self._annotation_count_stream])

    def overlay(
        self,
        indicators=True,
        editor=True,
        range_style=None,
        point_style=None,
        edit_range_style=None,
        edit_point_style=None,
        highlight=None,
    ) -> hv.Overlay:
        if range_style is None:
            range_style = Indicator.range_style
        if point_style is None:
            point_style = Indicator.point_style
        if edit_range_style is None:
            edit_range_style = Indicator.edit_range_style
        if edit_point_style is None:
            edit_point_style = Indicator.edit_point_style
        if highlight is None:
            highlight = Indicator.indicator_highlight

        highlighters = {opt: self.selected_dim_expr(v[0], v[1]) for opt, v in highlight.items()}
        indicator_style = Indicator.indicator_style(range_style, point_style, highlighters)
        region_style = Indicator.region_style(edit_range_style, edit_point_style)

        layers = []
        active_tools = []
        if "range" in self.region_types or self.region_types == "point":
            active_tools += ["box_select"]
        elif self.region_types == "point-point":
            active_tools += ["tap"]
        layers.append(self._element.opts(tools=self.edit_tools, active_tools=active_tools))

        if indicators:
            layers.append(self.indicators().opts(*indicator_style))
        if editor:
            layers.append(self.region_editor().opts(*region_style))
        return hv.Overlay(layers).collate()

    @property
    def static_indicators(self):
        data = self.annotator.get_dataframe(dims=self.kdims)
        fields_labels = self.annotator.all_fields
        region_labels = [k for k in data.columns if k not in fields_labels]

        indicator_kwargs = {
            "data": data,
            "region_labels": region_labels,
            "fields_labels": fields_labels,
            "invert_axes": False,  # Not yet handled
        }

        if self.region_types == "range":
            indicator = Indicator.ranges_1d(**indicator_kwargs)
        elif self.region_types == "range-range":
            indicator = Indicator.ranges_2d(**indicator_kwargs)
        elif self.region_types == "point":
            indicator = Indicator.points_1d(**indicator_kwargs)
        elif self.region_types == "point-point":
            indicator = Indicator.points_2d(**indicator_kwargs)
        else:
            msg = f"{self.region_types} not implemented"
            raise NotImplementedError(msg)

        return indicator

    def selected_dim_expr(self, selected_value, non_selected_value):
        self._selected_values.append(selected_value)
        self._selected_options.append({i: selected_value for i in self.annotator.selected_indices})
        index_name = (
            "id"
            if self.annotator.annotation_table._field_df.index.name is None
            else self.annotator.annotation_table._field_df.index.name
        )
        return hv.dim(index_name).categorize(
            self._selected_options[-1], default=non_selected_value
        )

    @property
    def dim_expr(self):
        return self._selection_info["dim_expr"]

    def show_region(self):
        kdims = list(self.kdims)
        region = {k: v for k, v in self.annotator._region.items() if k in self.kdims}

        if not region:
            return

        if self.region_types == "range":
            value = region[kdims[0]]
            bounds = (value[0], 0, value[1], 1)
        elif self.region_types == "range-range":
            bounds = (
                region[kdims[0]][0],
                region[kdims[1]][0],
                region[kdims[0]][1],
                region[kdims[1]][1],
            )
        elif self.region_types == "point":
            value = region[kdims[0]]
            bounds = (value, 0, value, 1)
        else:
            bounds = False

        if bounds:
            self._edit_streams[0].event(bounds=bounds)


class Annotator(AnnotatorInterface):
    """
    An annotator displays the contents of an AnnotationTable and
    provides the means to manipulate view the corresponding contents,
    add new annotations and update existing annotations.
    """

    def __init__(self, spec: dict, **params):
        """
        The spec argument must be an element or a dictionary of kdim dtypes
        """

        self._displays = {}

        super().__init__(
            spec if isinstance(spec, dict) else self._infer_kdim_dtypes(spec),
            **params,
        )

        self._selection_enabled = True
        self._editable_enabled = True

    @classmethod
    def _infer_kdim_dtypes(self, element: hv.Element) -> dict:
        # Remove?
        return AnnotationDisplay._infer_kdim_dtypes(element)

    def _create_annotation_element(self, element_key: tuple[str, ...]) -> AnnotationDisplay:
        for key in element_key:
            if key not in self.spec:
                msg = f"Dimension {key!r} not in spec"
                raise ValueError(msg)
        return AnnotationDisplay(self, kdims=list(element_key))

    def get_element(self, *kdims: str | hv.Dimension) -> hv.DynamicMap:
        return self.get_display(*kdims).element

    def get_display(self, *kdims: str | hv.Dimension) -> AnnotationDisplay:
        element_key = tuple(map(str, kdims))
        if element_key not in self._displays:
            self._displays[element_key] = self._create_annotation_element(element_key)
        return self._displays[element_key]

    def __mul__(self, other: hv.Element) -> hv.Overlay:
        return other * self.get_element(*other.kdims)

    def __rmul__(self, other: hv.Element) -> hv.Overlay:
        return self.__mul__(other)

    def refresh(self, clear=False) -> None:
        for v in self._displays.values():
            hv.streams.Stream.trigger([v._annotation_count_stream])
            if clear:
                v.clear_indicated_region()
            v.show_region()

    def set_annotation_table(self, annotation_table):
        self.select_by_index()
        for v in self._displays.values():
            v.select_by_index()
        self.clear_regions()
        super().set_annotation_table(annotation_table)
        self.refresh(clear=True)

    def clear_edits(self):
        super().clear_edits()
        self.clear_indicated_region()

    def add_annotation(self, **fields):
        super().add_annotation(**fields)
        self.refresh(clear=True)

    def update_annotation_region(self, index):
        super().update_annotation_region(index)
        self.refresh()

    def update_annotation_fields(self, index, **fields):
        super().update_annotation_fields(index, **fields)
        self.refresh()

    def delete_annotation(self, index):
        super().delete_annotation(index)
        self.refresh()

    def delete_annotations(self, *indices):
        if not indices:
            msg = "At least one index must be specified to delete annotations"
            raise ValueError(msg)
        for index in indices:
            super().delete_annotation(index)
        self.refresh()

    def select_by_index(self, *inds):
        "Set the selection state by the indices i.e. primary key values"

        for v in self._displays.values():
            if not v.selection_enabled:
                inds = ()

            for d, val in zip(v._selected_options, v._selected_values):
                d.clear()
                for ind in inds:
                    d[ind] = val
        super().select_by_index(*inds)
        self.refresh()

    def define_annotations(self, data: pd.DataFrame, **kwargs) -> None:
        super().define_annotations(data, **kwargs)
        self.refresh(clear=True)

    def revert_to_snapshot(self):
        super().revert_to_snapshot()
        self.refresh()

    def set_regions(self, **items):
        super().set_regions(**items)
        self.refresh()

    @property
    def selection_enabled(self) -> bool:
        return self._selection_enabled

    @selection_enabled.setter
    def selection_enabled(self, enabled: bool) -> None:
        for v in self._displays.values():
            v.selection_enabled = enabled

        if not enabled:
            self.select_by_index()

    @property
    def editable_enabled(self) -> bool:
        return self._editable_enabled

    @editable_enabled.setter
    def editable_enabled(self, enabled: bool) -> None:
        for v in self._displays.values():
            v.editable_enabled = enabled
